fecdec_filter: Avoid potentially expensive pointer subtraction.
[paraslash.git] / string.c
index eb8b12ece58ecd03e1507932751d1961de0084be..f51f3574f0cf372bdd8febb5ac517c45e3d92d59 100644 (file)
--- a/string.c
+++ b/string.c
@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2008 Andre Noll <maan@systemlinux.org>
+ * Copyright (C) 2004-2009 Andre Noll <maan@systemlinux.org>
  *
  * Licensed under the GPL v2. For licencing details see COPYING.
  */
@@ -33,7 +33,7 @@
 __must_check __malloc void *para_realloc(void *p, size_t size)
 {
        /*
-        * No need to check for NULL pointers: If p is NULL, the  call
+        * No need to check for NULL pointers: If p is NULL, the call
         * to realloc is equivalent to malloc(size)
         */
        assert(size);
@@ -59,11 +59,12 @@ __must_check __malloc void *para_realloc(void *p, size_t size)
  */
 __must_check __malloc void *para_malloc(size_t size)
 {
-       assert(size);
-       void *p = malloc(size);
+       void *p;
 
+       assert(size);
+       p = malloc(size);
        if (!p) {
-               PARA_EMERG_LOG("malloc failed (size = %zu),  aborting\n",
+               PARA_EMERG_LOG("malloc failed (size = %zu), aborting\n",
                        size);
                exit(EXIT_FAILURE);
        }
@@ -91,14 +92,14 @@ __must_check __malloc void *para_calloc(size_t size)
 }
 
 /**
- * Pparaslash's version of strdup().
+ * Paraslash's version of strdup().
  *
  * \param s The string to be duplicated.
  *
  * A wrapper for strdup(3). It calls \p exit(EXIT_FAILURE) on errors, i.e.
  * there is no need to check the return value in the caller.
  *
- * \return A pointer to the duplicated string. If \p s was the NULL pointer,
+ * \return A pointer to the duplicated string. If \a s was the \p NULL pointer,
  * an pointer to an empty string is returned.
  *
  * \sa strdup(3)
@@ -217,12 +218,13 @@ __must_check const char *para_basename(const char *name)
  *
  * \param buf The string to be chopped.
  *
* Replace the last character in \p buf by zero if it is euqal to
  * the newline character.
  */
 void chop(char *buf)
 {
        int n = strlen(buf);
+
        if (!n)
                return;
        if (buf[n - 1] == '\n')
@@ -233,7 +235,7 @@ void chop(char *buf)
  * Get a random filename.
  *
  * This is by no means a secure way to create temporary files in a hostile
- * direcory like \p /tmp. However, it is OK to use for temp files, fifos,
+ * directory like \p /tmp. However, it is OK to use for temp files, fifos,
  * sockets that are created in ~/.paraslash. Result must be freed by the
  * caller.
  *
@@ -253,8 +255,8 @@ __must_check __malloc char *para_tmpname(void)
 /**
  * Get the logname of the current user.
  *
- * \return A dynammically allocated string that must be freed by the caller. On
- * errors, the string "unknown user" is returned, i.e. this function never
+ * \return A dynamically allocated string that must be freed by the caller. On
+ * errors, the string "unknown_user" is returned, i.e. this function never
  * returns \p NULL.
  *
  * \sa getpwuid(3).
@@ -268,7 +270,7 @@ __must_check __malloc char *para_logname(void)
 /**
  * Get the home directory of the current user.
  *
- * \return A dynammically allocated string that must be freed by the caller. If
+ * \return A dynamically allocated string that must be freed by the caller. If
  * the home directory could not be found, this function returns "/tmp".
  */
 __must_check __malloc char *para_homedir(void)
@@ -284,16 +286,16 @@ __must_check __malloc char *para_homedir(void)
  * \param argv_ptr Pointer to the list of substrings.
  * \param delim Delimiter.
  *
- * This function modifies \a args by replacing each occurance of \a delim by
+ * This function modifies \a args by replacing each occurrence of \a delim by
  * zero. A \p NULL-terminated array of pointers to char* is allocated dynamically
  * and these pointers are initialized to point to the broken-up substrings
  * within \a args. A pointer to this array is returned via \a argv_ptr.
  *
  * \return The number of substrings found in \a args.
  */
-__must_check unsigned split_args(char *args, char *** const argv_ptr, const char *delim)
+unsigned split_args(char *args, char *** const argv_ptr, const char *delim)
 {
-       char *p = args;
+       char *p;
        char **argv;
        size_t n = 0, i, j;
 
@@ -327,31 +329,10 @@ __must_check unsigned split_args(char *args, char *** const argv_ptr, const char
        return n;
 }
 
-/**
- * Ensure that file descriptors 0, 1, and 2 are valid.
- *
- * Common approach that opens /dev/null until it gets a file descriptor greater
- * than two.
- *
- * \sa okir's Black Hats Manual.
- */
-void valid_fd_012(void)
-{
-       while (1) {
-               int fd = open("/dev/null", O_RDWR);
-               if (fd < 0)
-                       exit(EXIT_FAILURE);
-               if (fd > 2) {
-                       close(fd);
-                       break;
-               }
-       }
-}
-
 /**
  * Get the own hostname.
  *
- * \return A dynammically allocated string containing the hostname.
+ * \return A dynamically allocated string containing the hostname.
  *
  * \sa uname(2).
  */
@@ -428,7 +409,7 @@ static int for_each_complete_line(enum for_each_line_modes mode, char *buf,
 /**
  * Call a custom function for each complete line.
  *
- * \param buf The buffer containing data seperated by newlines.
+ * \param buf The buffer containing data separated by newlines.
  * \param size The number of bytes in \a buf.
  * \param line_handler The custom function.
  * \param private_data Pointer passed to \a line_handler.
@@ -477,14 +458,22 @@ int for_each_line_ro(char *buf, size_t size, line_handler_t *line_handler,
 }
 
 /**
- * Safely print into a buffer at a given offset
+ * Safely print into a buffer at a given offset.
  *
  * \param b Determines the buffer, its size, and the offset.
  * \param fmt The format string.
  *
  * This function prints into the buffer given by \a b at the offset which is
  * also given by \a b. If there is not enough space to hold the result, the
- * buffer size is doubled until the underlying call to vsnprintf() succeeds.
+ * buffer size is doubled until the underlying call to vsnprintf() succeeds
+ * or the size of the buffer exceeds the maximal size specified in \a b.
+ *
+ * In the latter case the unmodified \a buf and \a offset values as well as the
+ * private_data pointer of \a b are passed to the \a max_size_handler of \a b.
+ * If this function succeeds, i.e. returns a non-negative value, the offset of
+ * \a b is reset to zero and the given data is written to the beginning of the
+ * buffer.
+ *
  * Upon return, the offset of \a b is adjusted accordingly so that subsequent
  * calls to this function append data to what is already contained in the
  * buffer.
@@ -493,7 +482,7 @@ int for_each_line_ro(char *buf, size_t size, line_handler_t *line_handler,
  * initial buffer is allocated.
  *
  * \return The number of bytes printed into the buffer (not including the
- * therminating \p NULL byte).
+ * terminating \p NULL byte).
  *
  * \sa make_message(), vsnprintf(3).
  */
@@ -505,26 +494,35 @@ __printf_2_3 int para_printf(struct para_buffer *b, const char *fmt, ...)
                b->buf = para_malloc(128);
                b->size = 128;
                b->offset = 0;
-       } else if (b->size <= b->offset + 1) {
-               b->size *= 2;
-               b->buf = para_realloc(b->buf, b->size);
        }
        while (1) {
                char *p = b->buf + b->offset;
                size_t size = b->size - b->offset;
                va_list ap;
-               va_start(ap, fmt);
-               ret = vsnprintf(p, size, fmt, ap);
-               va_end(ap);
-               if (ret > -1 && ret < size) { /* success */
-                       b->offset += ret;
-                       break;
+               if (size) {
+                       va_start(ap, fmt);
+                       ret = vsnprintf(p, size, fmt, ap);
+                       va_end(ap);
+                       if (ret > -1 && ret < size) { /* success */
+                               b->offset += ret;
+                               return ret;
+                       }
+               }
+               /* check if we may grow the buffer */
+               if (!b->max_size || 2 * b->size < b->max_size) { /* yes */
+                       /* try again with more space */
+                       b->size *= 2;
+                       b->buf = para_realloc(b->buf, b->size);
+                       continue;
                }
-               /* try again with more space */
-               b->size *= 2;
-               b->buf = para_realloc(b->buf, b->size);
+               /* can't grow buffer */
+               if (!b->offset || !b->max_size_handler) /* message too large */
+                       return -ERRNO_TO_PARA_ERROR(ENOSPC);
+               ret = b->max_size_handler(b->buf, b->offset, b->private_data);
+               if (ret < 0)
+                       return ret;
+               b->offset = 0;
        }
-       return ret;
 }
 
 /** \cond LLONG_MAX and LLONG_LIN might not be defined. */
@@ -589,3 +587,34 @@ int para_atoi32(const char *str, int32_t *value)
        *value = tmp;
        return 1;
 }
+
+static inline int loglevel_equal(const char *arg, const char * const ll)
+{
+       return !strncasecmp(arg, ll, strlen(ll));
+}
+
+/**
+ * Compute the loglevel number from its name.
+ *
+ * \param txt The name of the loglevel (debug, info, ...).
+ *
+ * \return The numeric representation of the loglevel name.
+ */
+int get_loglevel_by_name(const char *txt)
+{
+       if (loglevel_equal(txt, "debug"))
+               return LL_DEBUG;
+       if (loglevel_equal(txt, "info"))
+               return LL_INFO;
+       if (loglevel_equal(txt, "notice"))
+               return LL_NOTICE;
+       if (loglevel_equal(txt, "warning"))
+               return LL_WARNING;
+       if (loglevel_equal(txt, "error"))
+               return LL_ERROR;
+       if (loglevel_equal(txt, "crit"))
+               return LL_CRIT;
+       if (loglevel_equal(txt, "emerg"))
+               return LL_EMERG;
+       return -1;
+}
@@ -217,12 +218,13 @@ __must_check const char *para_basename(const char *name)
  * \param buf The string to be chopped.
  *
- * Replace the last character in \p buf by zero if it is euqal to
+ * Replace the last character in \p buf by zero if it is equal to
  * the newline character.
  */